#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Date    : 2019-12-13 09:30:08
# @Author  : Chenghao Mou (chengham@isi.edu)

# pylint: disable=unused-wildcard-import

from typing import *
from loguru import logger

LABEL2INT = {
    "anli": {
        "1": 0,
        "2": 1,
    },
    "hellaswag": {
        "0": 0,
        "1": 1,
        "2": 2,
        "3": 3,
    },
    "piqa": {
        "0": 0,
        "1": 1,
    },
    "siqa": {
        "1": 0,
        "2": 1,
        "3": 2
    },
    "codah": {
        "0": 0,
        "1": 1,
        "2": 2,
        "3": 3
    },
    "cqa": {
        "A": 0,
        "B": 1,
        "C": 2,
        "D": 3,
        "E": 4
    },
    "cosmosqa": {
        "0": 0,
        "1": 1,
        "2": 2,
        "3": 3,
    },

}

INT2LABEL = {
    key: {
        i: l for l, i in LABEL2INT[key].items()
    }
    for key in LABEL2INT
}


TEMPLATES = {}
__all__ = ["LABEL2INT", "INT2LABEL", "TEMPLATES"]


def template(func):
    global __all__
    global TEMPLATES
    __all__.append(func.__name__)
    TEMPLATES[func.__name__] = func

    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)
    return wrapper


@template
def template_anli(datum: Dict, label2int: Dict[str, int]) -> Dict[str, Any]:

    template: Dict[str, Any] = {
        "text": [
                [datum["obs1"], datum["hyp1"], datum["obs2"]],
                [datum["obs1"], datum["hyp2"], datum["obs2"]],
        ],
        "label": None if "label" not in datum else label2int[str(datum["label"])],
        "image": None,
        "token_type_id": [0, 1, 0],
        "attention": [1, 1, 1]
    }

    assert template['label'] is not None, "anli"

    return template


@template
def template_hellaswag(datum: Dict, label2int: Dict[str, int]) -> Dict[str, Any]:

    template: Dict[str, Any] = {
        "text": [
            [datum["ctx"], x] for x in datum["ending_options"]
        ],
        "label": None if "label" not in datum else label2int[str(datum['label'])],
        "image": None,
        "token_type_id": [0, 1],
        "attention": [1, 1]
    }
    assert template['label'] is not None, "hellaswag"

    return template


@template
def template_piqa(datum: Dict, label2int: Dict[str, int]) -> Dict[str, Any]:

    template = {
        "text": [
                [datum["goal"], datum["sol1"]],
                [datum["goal"], datum["sol2"]],
        ],
        "label": None if "label" not in datum else label2int[str(datum['label'])],
        "image": None,
        "token_type_id": [0, 1],
        "attention": [1, 1]
    }

    assert template['label'] is not None, "piqa"

    return template


@template
def template_siqa(datum: Dict, label2int: Dict[str, int]) -> Dict[str, Any]:

    template = {
        "text": [
                [datum["context"], datum["question"], datum["answerA"]],
                [datum["context"], datum["question"], datum["answerB"]],
                [datum["context"], datum["question"], datum["answerC"]],
        ],
        "label": None if "label" not in datum else label2int[str(datum['label'])],
        "image": None,
        "token_type_id": [0, 0, 1],
        "attention": [1, 1, 1]
    }

    assert template['label'] is not None, "siqa"

    return template


@template
def template_cqa(datum: Dict, label2int: Dict[str, int]) -> Dict[str, Any]:

    template = {
        "text": [
                [datum["question"]["stem"], x['text']] for x in datum["question"]["choices"]
        ],
        "label": None if "answerKey" not in datum else label2int[str(datum["answerKey"])],
        "image": None,
        "token_type_id": [0,  1],
        "attention": [1, 1]
    }

    assert template['label'] is not None, "cqa"

    return template


@template
def template_cosmosqa(datum: Dict, label2int: Dict[str, int]) -> Dict[str, Any]:

    template = {
        "text": [
                [datum["context"], datum['question'], datum["answer0"]],
                [datum["context"], datum['question'], datum["answer1"]],
                [datum["context"], datum['question'], datum["answer2"]],
                [datum["context"], datum['question'], datum["answer3"]],

        ],
        "label": None if "label" not in datum else label2int[str(datum["label"])],
        "image": None,
        "token_type_id": [0, 0, 1],
        "attention": [1, 1, 1]
    }

    assert template['label'] is not None, "cqa"

    return template


@template
def template_codah(datum: Dict, label2int: Dict[str, int]) -> Dict[str, Any]:

    result = {
        # "category": datum['category'],
        "text": [
            [datum['prompt'], datum['answer0']],
            [datum['prompt'], datum['answer1']],
            [datum['prompt'], datum['answer2']],
            [datum['prompt'], datum['answer3']]
        ],
        "image": None,
        "label": None if datum['label'] is None else label2int[str(datum['label'])],
        "token_type_id": [0, 1],
        "attention": [1, 1],
    }

    return result


@template
def template_smthsmth(datum: Dict, label2int: Dict[str, int]) -> Dict[str, Any]:
    template = {
        "text": [
                [datum["label"]]
        ],
        "label": None,
        "image": datum["id"],
        "token_type_id": [0],
        "attention": [1]
    }

    return template



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Date    : 2019-12-13 10:43:56
# @Author  : Chenghao Mou(chengham@isi.edu)

# pylint: disable=no-member
# pylint: disable=unused-wildcard-import
# pylint: disable=invalid-sequence-index
# pylint: disable=not-callable

import os
import random
import json
from typing import *
from pathlib import Path
from dataclasses import dataclass
from itertools import zip_longest

from .templates import *
from .renderers import *

import torch
from torch.utils.data import Dataset, Sampler
from torch.nn.utils.rnn import pad_sequence
from transformers import PreTrainedTokenizer
import pandas as pd


class TokenBasedSampler(Sampler):

    def __init__(self, data_source, batch_size=128):
        self.data_source = data_source
        self.batch_size = batch_size

    def __iter__(self):

        indices: List[List[int]] = [[]]
        curr_max: int = 0
        for i, x in enumerate(self.data_source):
            num, max_len = len(x['input_id']), max(map(len, x['input_id']))
            curr_max = max(max_len, curr_max)
            l = ((len(indices[-1]) + 1) * num) * curr_max
            if l > self.batch_size and indices[-1]:
                indices.append([i])
                curr_max = max_len
            else:
                indices[-1].append(i)

        return iter(indices)


@dataclass
class BatchTool:
    """Batching utility class for text

    tokenizer: PreTrainedTokenizer from transformers, used for padding
    max_seq_len: Maximum length of each input sequence;
    mlm: whether to use MLM in the batch;
    mlm_probability: the probability of masked language model;
    source: str identifier for the provenance of the batch, useful for multitasking;

    """

    tokenizer: PreTrainedTokenizer
    max_seq_len: int = 128
    mlm: bool = False
    mlm_probability: float = 0.15
    source: str = ""

    def collate_fn(self, samples: List[Dict[str, Any]]) -> Dict[str, Any]:

        template: Dict = {
            "input_ids": [],
            "labels": [],
            "attentions": [],
            "images": [],
            "token_type_ids": [],
            "source": self.source
        }

        for sample in samples:

            template["input_ids"].append(torch.LongTensor(sample["input_id"]).transpose(0, 1))
            template["attentions"].append(torch.LongTensor(sample["attention"]).transpose(0, 1))
            template["token_type_ids"].append(torch.LongTensor(sample["token_type_id"]).transpose(0, 1))
            if sample["image"] is not None:
                template["images"].append(sample["image"])
            if sample["label"] is not None:
                template["labels"].append(sample["label"])

        template["input_ids"]: torch.tensor = pad_sequence(template["input_ids"], batch_first=True).transpose(1, 2)
        template["attentions"]: torch.tensor = pad_sequence(template["attentions"], batch_first=True).transpose(1, 2)
        template["token_type_ids"]: torch.tensor = pad_sequence(template["token_type_ids"], batch_first=True).transpose(1, 2)

        template["input_ids"]: torch.tensor = template["input_ids"][..., :self.max_seq_len]
        template["attentions"]: torch.tensor = template["attentions"][..., :self.max_seq_len]
        template["token_type_ids"]: torch.tensor = template["token_type_ids"][..., :self.max_seq_len]

        template["images"] = torch.stack(template["images"]) if template["images"] else None

        if template["labels"] != []:
            template["labels"] = torch.LongTensor(template["labels"])
        if self.mlm == True:
            batch_size, num_choice, seq_length = template["input_ids"].shape
            template["input_ids"], template["labels"] = self.mask_tokens(
                template["input_ids"].reshape(-1, seq_length))
            template["input_ids"] = template["input_ids"].reshape(batch_size, num_choice, seq_length)
            template["labels"] = template["labels"].reshape(batch_size, num_choice, seq_length)

        return template

    def mask_tokens(self, input_ids):
        labels = input_ids.clone()
        probability_matrix = torch.full(labels.shape, self.mlm_probability)
        special_tokens_mask = [self.tokenizer.get_special_tokens_mask(val, already_has_special_tokens=True) for val in labels.tolist()]
        probability_matrix.masked_fill_(torch.tensor(special_tokens_mask, dtype=torch.bool), value=0.0)
        masked_indices = torch.bernoulli(probability_matrix).bool()
        labels[~masked_indices] = -1  # We only compute loss on masked tokens

        # 80% of the time, we replace masked input tokens with tokenizer.mask_token ([MASK])
        indices_replaced = torch.bernoulli(torch.full(labels.shape, 0.8)).bool() & masked_indices
        input_ids[indices_replaced] = self.tokenizer.mask_token_id

        # 10% of the time, we replace masked input tokens with random word
        indices_random = torch.bernoulli(torch.full(labels.shape, 0.5)).bool() & masked_indices & ~indices_replaced
        random_words = torch.randint(len(self.tokenizer), labels.shape, dtype=torch.long)
        input_ids[indices_random] = random_words[indices_random]

        # The rest of the time (10% of the time) we keep the masked input tokens unchanged
        return input_ids, labels

    @classmethod
    def uncollate_fn(cls, samples: List):

        assert len(samples) == 1

        return samples[0]


__all__ = ["BatchTool", "TokenBasedSampler"]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Date    : 2019-11-17 18:09:14
# @Author  : Chenghao Mou (mouchenghao@gmail.com)
# @Link    : link
# @Version : 1.0.0

# pylint: disable=unused-wildcard-import
# pylint: disable=no-member
# pylint: disable=not-callable

import os
import random
import json
from typing import *
from pathlib import Path
from dataclasses import dataclass

from .templates import *
from .renderers import *

import pandas as pd
from torch.utils.data import Dataset
from tqdm import tqdm
from loguru import logger


class MultiModalDataset(Dataset):

    def __init__(self, df: pd.DataFrame, template: Callable, renderers: List[Callable]):
        df = df.apply(template, axis=1, result_type='expand')
        for renderer in renderers:
            df = df.apply(renderer, axis=1, result_type='expand')
        self.data: List[Dict] = df.to_dict('records')

    def __len__(self):

        return len(self.data)

    def __getitem__(self, index):

        return self.data[index]


class MultiTaskDataset(Dataset):

    def __init__(self, dataloaders, shuffle: bool = True):

        self.data: List = []

        for loader in tqdm(dataloaders):
            for batch in tqdm(loader, leave=False):
                self.data.append(batch)

        if shuffle:
            random.shuffle(self.data)

    def __len__(self):
        return len(self.data)

    def __getitem__(self, index):
        return self.data[index]


__all__ = ["MultiTaskDataset", "MultiModalDataset"]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Date    : 2019-12-13 09:48:30
# @Author  : Chenghao Mou (chengham@isi.edu)

# pylint: disable=no-member
# pylint: disable=unused-wildcard-import


import os
from typing import *
from pathlib import Path

from .transforms_video import *

import torchvision
import av
from transformers import PreTrainedTokenizer
from loguru import logger

SCALE = 1.1


def renderer_text(datum: Dict[str, Any], tokenizer: PreTrainedTokenizer, task_token: str = None) -> Dict[str, Any]:

    assert getattr(tokenizer, "cls_token") is not None, "cls token not found"
    assert getattr(tokenizer, "pad_token") is not None, "pad token not found"
    assert getattr(tokenizer, "mask_token") is not None, "mask token not found"

    task_token = task_token if task_token is not None else tokenizer.cls_token

    template = {
        "input_id": [[] for _ in range(len(datum["text"]))],
        "label": datum["label"],
        "attention": [[] for _ in range(len(datum["text"]))],
        "token_type_id": [[] for _ in range(len(datum["text"]))],
        "image": datum["image"],
    }

    max_seq_len = 0

    for i, group in enumerate(datum["text"]):

        for j, (attn, token_type, sentence) in enumerate(zip(datum["attention"], datum["token_type_id"], group)):
            tokens = tokenizer.tokenize(sentence)
            if j == 0:
                tokens = [task_token] + tokens
            tokens = tokens + [tokenizer.pad_token]

            template["attention"][i].extend(attn for _ in tokens)
            template["token_type_id"][i].extend(token_type for _ in tokens)
            template["input_id"][i].extend(tokenizer.convert_tokens_to_ids(tokens))

        max_seq_len = max(max_seq_len, len(template["input_id"][i]))

    for j in range(len(template["input_id"])):
        while len(template["input_id"][j]) < max_seq_len:
            template["input_id"][j].append(tokenizer.pad_token_id)
            template["attention"][j].append(0)
            template["token_type_id"][j].append(0)

    return template


def renderer_video(datum: Dict[str, Any],
                   data_dir: Union[Path, str],
                   nframe: int = 72,
                   nclip: int = 1,
                   width: int = 64,
                   dstep_size: int = 1):

    upscale_size = int(width * SCALE)
    transform_pre = ComposeMix([
        [Scale(upscale_size), "img"],
        [RandomCropVideo(width), "vid"],
    ])
    transform_post = ComposeMix([
        [torchvision.transforms.ToTensor(), "img"],
    ])

    path = os.path.join(data_dir, datum["image"] + ".webm")
    reader = av.open(path)

    try:
        imgs = []
        imgs = [f.to_rgb().to_ndarray() for f in reader.decode(video=0)]
    except (RuntimeError, ZeroDivisionError) as exception:
        logger.debug('{}: WEBM reader cannot open {}. Empty list returned.'.format(type(exception).__name__, path))

    imgs = transform_pre(imgs)
    imgs = transform_post(imgs)

    num_frames = len(imgs)

    if nclip > -1:
        num_frames_necessary = nframe * nclip * dstep_size
    else:
        num_frames_necessary = num_frames

    offset = 0
    if num_frames_necessary < num_frames:
        # If there are more frames, then sample a starting offset.
        diff = (num_frames - num_frames_necessary)
        offset = np.random.randint(0, diff)

    imgs = imgs[offset: num_frames_necessary + offset: dstep_size]

    if len(imgs) < (nframe * nclip):
        imgs.extend([imgs[-1]] * ((nframe * nclip) - len(imgs)))

    data = torch.stack(imgs)
    data = data.permute(1, 0, 2, 3)

    datum["image"] = data

    return datum


__all__ = ["renderer_video", "renderer_text"]



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="textbook",
    version_format='{tag}.dev{commitcount}+{gitsha}',
    setup_requires=['setuptools-git-version'],
    author="Chenghao",
    python_requires='>=3.7.0',
    author_email="mouchenghao@gmail.com",
    description="Text classification datasets",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/ChenghaoMou/textbook",
    packages=setuptools.find_packages(),
    install_requires=[
        'torch',
        'transformers',
        'torchvision',
        'av',
        'pandas',
        'tqdm',
        'loguru',
        'numpy',
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: MacOS :: MacOS X",
    ],
)
